#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2025 Carlos M. Fonseca <cmfonsec@dei.uc.pt>
#
# SPDX-License-Identifier: Apache-2.0

import random

from collections.abc import Iterable
from utils.types import Solution, Neighbourhood, Move, Increment

from api.base import Neighbourhood, Move
from api.neighborhood_destr_pop import PopMove

# ------------------------------- Neighbourhood ------------------------------

class AddNeighbourhood(Neighbourhood):
    def moves(self, solution: Solution) -> Iterable[Move]:
        assert self.problem == solution.problem
        dd = [(job,self.problem.due_dates[job]) for job in solution.not_scheduled]
        for job, _ in sorted(dd, key=lambda x: x[1]):
            yield AddMove(self, job)

    def random_moves_without_replacement(self, solution: Solution) -> Iterable[Move]:
        assert self.problem == solution.problem
        if not solution.is_feasible():
            for i in random.sample(solution.not_scheduled, k=len(solution.not_scheduled)):
                yield AddMove(self, i)



# ----------------------------------- Move -----------------------------------

class AddMove(Move):
    def __init__(self, neighbourhood, job, increment=None):
        self.neighbourhood = neighbourhood
        # i is the job to add
        self.job = job
        self.increment = None
        self._cached_hash = hash(self.job)

    def __str__(self):
        return "add: %d, incr: %d" % (self.job, self.increment)

    def __eq__(self, other):
        if not isinstance(other, AddMove):
            return False
        return self._cached_hash == other._cached_hash

    def __hash__(self):
        return self._cached_hash

    def invert_move(self) -> PopMove:
        # Invert the move
        return PopMove(self.neighbourhood, -self.increment)
        # return RemoveMove(self.neighbourhood, self.job, -self.increment) 

    def apply(self, solution: Solution) -> Solution:
        #Update completion time
        job_i = self.job
        index = len(solution.sequence)
        if index >0:
            solution.completion_time[index] = solution.completion_time[index-1] + solution.problem.processing_times[job_i]
        else:
            solution.completion_time[index] = solution.problem.processing_times[job_i]
        solution.tardiness[index] = max(0, solution.completion_time[index] - solution.problem.due_dates[job_i])
        # Update lower bound
        solution.lb += self.increment if self.increment is not None else 0
        # Update value
        if len(solution.not_scheduled)!=0:
            solution.value = None
        else:
            solution.value = solution.lb
        # Update solution
        solution.sequence.append(job_i)
        solution.not_scheduled.remove(job_i)
        return solution

    def lower_bound_increment(self, solution: Solution) -> Increment:
        prob = solution.problem
        job_i = self.job
        index = len(solution.sequence)
        if index > 0:
            completion_time = solution.completion_time[index-1] + prob.processing_times[job_i]
        else:
            completion_time = prob.processing_times[job_i]
        tardiness = max(0, completion_time - prob.due_dates[job_i])
        self.increment = prob.weights[job_i] * tardiness
        return self.increment
    