
import math
import random

from collections.abc import Iterable
from utils.types import Solution, Neighbourhood, Move, Increment

from api.base import Neighbourhood, Move
# ------------------------------- Neighbourhood ------------------------------

class InsertNeighbourhood(Neighbourhood):
    def moves(self, solution: Solution) -> Iterable[Move]:
        assert self.problem == solution.problem
        # raise NotImplementedError("Not implemented yet")
        n = self.problem.n
        if solution.is_feasible():
            for ix in range(n):
                for jx in range(n):
                   if ix!=jx:
                       yield InsertMove(self, ix, jx)


    def random_moves_without_replacement(self, solution: Solution) -> Iterable[Move]:
        assert self.problem == solution.problem
        n = self.problem.n
        if solution.is_feasible():
            for ix in random.sample(range(n), k=n):
                for jx in random.sample(range(n), k=n-1):
                    if ix!=jx:
                        yield InsertMove(self, ix, jx)

    def random_move(self, solution: Solution) -> Move:
        assert self.problem == solution.problem
        n = self.problem.n
        if solution.is_feasible():
            ix = random.randint(0, n-1)
            jx = random.randint(0, n-1)
            while ix==jx:
                jx = random.randint(0, n-1)
            return InsertMove(self, ix, jx)
        raise ValueError("Solution is not feasible, cannot generate random move.")

# ----------------------------------- Move -----------------------------------

class InsertMove(Move):
    def __init__(self, neighbourhood, ix, jx, increment=None):
        self.neighbourhood = neighbourhood        
        assert ix >= 0 and ix<=neighbourhood.problem.n-1
        assert jx >= 0 and jx<=neighbourhood.problem.n-1
        assert ix!=jx
        self.i = ix
        self.j = jx
        self.increment = increment
        self._cached_hash = hash((self.i, self.j))

    def __str__(self):
        return "insert %d -> %d; increment: %d" % (self.i, self.j, self.increment)
        
    def __eq__(self, other):
        if not isinstance(other, Move):
            return False
        return self._cached_hash == other._cached_hash

    def __hash__(self):
        return self._cached_hash

    def invert_move(self) -> Move:
        return InsertMove(self.neighbourhood, self.j, self.i, self.increment) 

    def apply(self, solution: Solution) -> Solution:
        # Update value
        if len(solution.not_scheduled)!=0:
            solution.value = None
        else:
            if self.increment is None:
                self.increment = self.objective_value_increment(solution)
            solution.value += self.increment 
            solution.lb = solution.value
        if self.i<self.j:
            # Update completion time   
            job_i = solution.sequence[self.i]
            for k in range(self.i, self.j):
                job_k = solution.sequence[k+1]
                if k-1<0:
                    solution.completion_time[k] = solution.problem.processing_times[job_k]
                else:
                    solution.completion_time[k] = solution.completion_time[k-1] + solution.problem.processing_times[job_k]
                solution.tardiness[k] = max(0, solution.completion_time[k] - solution.problem.due_dates[job_k])
            solution.completion_time[self.j] = solution.completion_time[self.j-1] + solution.problem.processing_times[job_i]
            solution.tardiness[self.j] = max(0, solution.completion_time[self.j] - solution.problem.due_dates[job_i])
            # Update solution
            tmp = solution.sequence[self.i]
            solution.sequence[self.i:self.j] = solution.sequence[self.i+1:self.j+1]
            solution.sequence[self.j] = tmp
        else:
            # Update completion time   
            job_i = solution.sequence[self.i]
            if self.j==0:
                solution.completion_time[self.j] = solution.problem.processing_times[job_i]
            else:
                solution.completion_time[self.j] = solution.completion_time[self.j-1] + solution.problem.processing_times[job_i]
            solution.tardiness[self.j] = max(0, solution.completion_time[self.j] - solution.problem.due_dates[job_i])
            for k in range(self.j+1, self.i+1):
                job_k = solution.sequence[k-1]
                solution.completion_time[k] = solution.completion_time[k-1] + solution.problem.processing_times[job_k]
                solution.tardiness[k] = max(0, solution.completion_time[k] - solution.problem.due_dates[job_k])
            # Update solution            
            solution.sequence[self.j+1:self.i+1] = solution.sequence[self.j:self.i]
            solution.sequence[self.j] = job_i
        return solution



    def objective_value_increment(self,  solution: Solution) -> Increment:
        prob = solution.problem
        job_i = solution.sequence[self.i]
        job_j = solution.sequence[self.j]
        increment = 0
        completion_time = 0

        if self.i<self.j:
            if self.i>0:
                completion_time = solution.completion_time[self.i-1]
            else:
                completion_time = 0
            for k in range(self.i, self.j):
                job_k = solution.sequence[k+1]
                completion_time += prob.processing_times[job_k]
                increment -= prob.weights[solution.sequence[k]] * solution.tardiness[k]
                increment += prob.weights[job_k] * max(0, completion_time - prob.due_dates[job_k])
            completion_time += prob.processing_times[job_i]
            increment -= prob.weights[job_j] * solution.tardiness[self.j]
            increment += prob.weights[job_i] * max(0, completion_time - prob.due_dates[job_i])
        else:
            if self.j>0:
                completion_time = solution.completion_time[self.j-1] + prob.processing_times[job_i]
            else:
                completion_time = 0 + prob.processing_times[job_i]
            increment -= prob.weights[job_j] * solution.tardiness[self.j]
            increment += prob.weights[job_i] * max(0, completion_time - prob.due_dates[job_i])
            for k in range(self.j+1, self.i+1):
                job_k = solution.sequence[k-1]
                completion_time += prob.processing_times[job_k]
                increment -= prob.weights[solution.sequence[k]] * solution.tardiness[k]
                increment += prob.weights[job_k] * max(0, completion_time - prob.due_dates[job_k])
            
        self.increment = increment
        return increment

