#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2025 Carlos M. Fonseca <cmfonsec@dei.uc.pt>
#
# SPDX-License-Identifier: Apache-2.0

import math
import random

from collections.abc import Iterable
from utils.types import Solution, Neighbourhood, Move, Increment

from api.base import Neighbourhood, Move

# ------------------------------- Neighbourhood ------------------------------

class SwapNeighbourhood(Neighbourhood):
    def moves(self, solution: Solution) -> Iterable[Move]:
        assert self.problem == solution.problem
        # raise NotImplementedError("Not implemented yet")
        n = self.problem.n
        if solution.is_feasible():
            for ix in range(n-1):
                yield SwapMove(self, ix)


    def random_moves_without_replacement(self, solution: Solution) -> Iterable[Move]:
        assert self.problem == solution.problem
        n = self.problem.n
        if solution.is_feasible():
            for ix in random.sample(range(n-1), k=n-1):
                yield SwapMove(self, ix)
    
    def random_move(self, solution: Solution) -> Move:
        assert self.problem == solution.problem
        n = self.problem.n
        if solution.is_feasible():
            ix = random.randint(0, n-2)
            return SwapMove(self, ix)
        raise ValueError("Solution is not feasible, cannot generate random move.")

# ----------------------------------- Move -----------------------------------

class SwapMove(Move):
    def __init__(self, neighbourhood, ix, increment=None):
        self.neighbourhood = neighbourhood
        # ix is the position to swap with the successor
        assert ix >= 0 and ix<neighbourhood.problem.n-1
        self.i = ix
        self.increment = increment
        self._cached_hash = hash(self.i)
        
    def __str__(self):
        return "swap %d; increment: %d" % (self.i, self.increment)

    def __repr__(self):
        return "swap %d; increment: %d" % (self.i, self.increment)

    def __eq__(self, other):
        if not isinstance(other, Move):
            return False
        return self._cached_hash == other._cached_hash

    def __hash__(self):
        return self._cached_hash

    def invert_move(self) -> Move:
        # In this case, the move is its own inverse
        return SwapMove(self.neighbourhood, self.i, self.increment) 

    def apply(self, solution: Solution) -> Solution:
        # Update value
        if len(solution.not_scheduled)!=0:
            solution.value = None
        else:
            if self.increment is None:
                self.increment = self.objective_value_increment(solution)
            solution.value += self.increment 
            solution.lb = solution.value
        # Update completion time   
        job_i = solution.sequence[self.i]
        job_j = solution.sequence[self.i+1] 
        if self.i>0:
            solution.completion_time[self.i] = solution.completion_time[self.i-1] + solution.problem.processing_times[job_j]
        else:
            solution.completion_time[self.i] = solution.problem.processing_times[job_j]        
        solution.tardiness[self.i] = max(0, solution.completion_time[self.i] - solution.problem.due_dates[job_j])
        solution.completion_time[self.i+1] = solution.completion_time[self.i] + solution.problem.processing_times[job_i]
        solution.tardiness[self.i+1] = max(0, solution.completion_time[self.i+1] - solution.problem.due_dates[job_i])
        # Update solution
        tmp = solution.sequence[self.i]
        solution.sequence[self.i] = solution.sequence[self.i+1]
        solution.sequence[self.i+1] = tmp
        return solution

    def objective_value_increment(self,  solution: Solution) -> Increment:
        prob = solution.problem
        job_i = solution.sequence[self.i]
        job_j = solution.sequence[self.i+1]
        increment = 0
        completion_time = 0
        if self.i>0:
            completion_time = solution.completion_time[self.i-1]
        increment -= prob.weights[job_i] * solution.tardiness[self.i]
        increment -=  prob.weights[job_j] * solution.tardiness[self.i+1]
        completion_time += prob.processing_times[job_j] 
        increment +=  prob.weights[job_j] * max(0, completion_time- prob.due_dates[job_j])
        completion_time += prob.processing_times[job_i]
        increment +=  prob.weights[job_i] * max(0, completion_time - prob.due_dates[job_i])

        self.increment = increment
        return increment
    

