#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2025 Carlos M. Fonseca <cmfonsec@dei.uc.pt>
#
# SPDX-License-Identifier: Apache-2.0

import random
import matplotlib.pyplot as plt  

from api.base import Neighbourhood
from api.neighborhood_constr_add import AddNeighbourhood
from api.neighborhood_destr_pop import PopMove
from api.neighborhood_local_swap import SwapNeighbourhood
from api.neighborhood_local_insert import InsertNeighbourhood
from api.neighborhood_local_union import InsertOrSwap


class Solution:
    def __init__(self, problem, sequence, not_scheduled, lb, value=None, completion_time=None, tardiness=None):
        self.problem = problem
        self.sequence = sequence
        self.not_scheduled = not_scheduled
        self.lb = lb
        self.value = value

        # Auxiliary variables to speed up incremental computations 
        self.completion_time = completion_time
        self.tardiness = tardiness

        self.c_nbhood = None
        self.l_nbhood = None

    def __str__(self):
        return " ".join(map(str, self.sequence))

    def copy(self):
        return self.__class__(self.problem,
            self.sequence.copy(),
            self.not_scheduled.copy(),
            self.lb, self.value,
            self.completion_time, self.tardiness)
    
    def __eq__(self, other):
        if not isinstance(other, Solution):
            return False
        assert self.problem == other.problem 
        assert self.sequence == other.sequence 
        assert self.not_scheduled == other.not_scheduled
        assert self.lb == other.lb        
        assert self.completion_time == other.completion_time
        assert self.tardiness == other.tardiness
        assert self.value == other.value
        return (self.problem == other.problem and
                self.sequence == other.sequence and
                self.not_scheduled == other.not_scheduled and
                self.lb == other.lb and
                self.value == other.value and
                self.completion_time == other.completion_time and
                self.tardiness == other.tardiness)
    

    @classmethod
    def empty_solution(cls, problem):
        return cls(problem, [], set(range(problem.n)), 0, None, [0]*problem.n, [0]*problem.n)

    @classmethod
    def random_solution(cls, problem):
        c = list(range(problem.n))
        random.shuffle(c)
        total_weighted_tardiness = 0
        completion_time = [0]*problem.n
        tardiness = [0]*problem.n
        for i in range(problem.n):
            completion_time[i] = completion_time[i-1] + problem.processing_times[c[i]]
            tardiness[i] = max(0,  completion_time[i] - problem.due_dates[c[i]])
            total_weighted_tardiness += problem.weights[c[i]] * tardiness[i]
        return cls(problem, c, set(), total_weighted_tardiness, total_weighted_tardiness, completion_time, tardiness)

    @classmethod
    def given_solution(cls, problem):
        sequence = [3, 0, 4, 2, 1]
        not_scheduled = set()
        return cls(problem, sequence, not_scheduled, 0, None, [0]*problem.n, [0]*problem.n)

    def problem_instance(self):
        return self.problem

    def is_feasible(self):
        return len(self.not_scheduled) == 0

    def objective_value(self):
        total_weighted_tardiness = 0
        completion_time = [0]*self.problem.n
        tardiness = [0]*self.problem.n
        if self.is_feasible():
            for i in range(self.problem.n):
                job_i = self.sequence[i]
                completion_time[i] = completion_time[i-1] + self.problem.processing_times[job_i]
                tardiness[i] = max(0, completion_time[i] - self.problem.due_dates[job_i] )
                total_weighted_tardiness += self.problem.weights[job_i] * tardiness[i] 
                        
            self.value = total_weighted_tardiness
            self.completion_time= completion_time
            self.tardiness = tardiness
            self.lb = total_weighted_tardiness
        else:
            self.value = None
        return self.value


    def lower_bound(self):
        total_weighted_tardiness = 0
        completion_time = [0]*self.problem.n
        tardiness = [0]*self.problem.n
        for i in range(len(self.sequence)):
            job_i = self.sequence[i]
            completion_time[i] = completion_time[i-1] + self.problem.processing_times[job_i]
            tardiness[i] = max(0,  completion_time[i] - self.problem.due_dates[job_i])
            total_weighted_tardiness += self.problem.weights[job_i] * tardiness[i] 
        self.completion_time= completion_time
        self.tardiness = tardiness
        self.lb = total_weighted_tardiness
        return self.lb

    def construction_neighbourhood(self):
        if self.c_nbhood is None:
            self.c_nbhood = AddNeighbourhood(self.problem)
        return self.c_nbhood

    def local_sub_neighbourhoods(self) -> list[Neighbourhood]:
        l_neighborhoods = []
        l_neighborhoods.append(SwapNeighbourhood(self.problem))
        l_neighborhoods.append(InsertNeighbourhood(self.problem))
        return l_neighborhoods
    
    def local_neighbourhood(self) -> Neighbourhood:
        if self.l_nbhood is None:
            #self.l_nbhood = InsertNeighbourhood(self.problem)
            self.l_nbhood = InsertOrSwap(self.problem)
        return self.l_nbhood

    def perturbation_neighbourhood(self) -> Neighbourhood:
        return InsertOrSwap(self.problem)

        