#!/usr/bin/env python3
#
# SPDX-FileCopyrightText: 2025 Carlos M. Fonseca <cmfonsec@dei.uc.pt>
#
# SPDX-License-Identifier: Apache-2.0

import roas.solvers as solvers
from api.problem import Problem
from api.solution import Solution
import sys
import math

import logging
from utils.log_config import setup_logging
setup_logging()
logger = logging.getLogger("main")

if __name__ == '__main__':
    if len(sys.argv)==2:
        with open(sys.argv[1], 'r') as f:
            prob = Problem.from_textio(f)
    else:
        raise ValueError("Usage: python3 main.py <problem_file>")
    if (prob):
        print(prob)
        sol = Solution.random_solution(prob)
        print("Random solution: {:.2f}".format(sol.objective_value()))
        print("Neighbourhood size:", len(list(sol.local_neighbourhood().moves(sol))))

        sol1 = solvers.greedy_construction(Solution.empty_solution(prob))        
        print("Greedy construction: {:.2f}".format(sol1.objective_value()))
        
        sol2 = solvers.greedy_construction_random_tie_breaking(Solution.empty_solution(prob))
        print("Greedy construction with random tie breaking: {:.2f}".format(sol2.objective_value()))

        sol3 = solvers.first_improvement(sol.copy())
        print("First improvement: {:.2f}".format(sol3.objective_value()))

        sol4 = solvers.best_improvement(sol.copy())
        print("Best improvement: {:.2f}".format(sol4.objective_value()))
        
        sol6 = solvers.grasp(Solution.empty_solution(prob), 5, 0.1, 42, solvers.first_improvement)
        print("GRASP: {:.2f}".format(sol6.objective_value()))
        
        sol7 = solvers.beam_search(Solution.empty_solution(prob), bw=4, FullRollout=True, bf=3)
        print("Beam search: {:.2f}".format(sol7.objective_value()))
        
        sol5 = solvers.basic_variable_neighbourhood_descent(sol.copy())
        print("Basic variable neighbourhood descent: {:.2f}".format(sol5.objective_value()))

        sol5b = solvers.basic_variable_neighbourhood_search(sol.copy(), time_budget=5, max_no_improv=math.inf)
        print("Basic variable neighbourhood search: {:.2f}".format(sol5b.objective_value()))

        sol8 = solvers.ils(sol.copy(), kickStrength=5, budget=5)
        print("Iterated Local Search: {:.2f}".format(sol8.objective_value()))
        
        sol9 = solvers.rls(sol.copy(), budget=5)
        print("Randomized Local Search: {:.2f}".format(sol9.objective_value()))

        sol99 = solvers.ls_with_side_walks(sol.copy(), budget=5)
        print("LS with side walks: {:.2f}".format(sol99.objective_value()))

        sol10 = solvers.sa(sol.copy(), budget=5, init_temp=10)
        print("Simulated Annealing: {:.2f}".format(sol10.objective_value()))

        sol11 = solvers.tabu_search(sol.copy(), tabu_interval = (5,10), time_budget=5, max_no_improv=100)
        print("Tabu Search: {:.2f}".format(sol11.objective_value()))




