import random, operator, time, math
import logging
logger = logging.getLogger("solvers")

from time import perf_counter

from utils.types import Solution, Value, Neighbourhood, Move, Increment

def construct_ant(solution: Solution,
                  alpha: float,
                  beta: float,
                  tau: dict[ComponentId, float]) -> Solution:
    while True:
        cs = []
        cszero = []
        p = []
        best = None
        for c in solution.add_moves():
            k = c.cid
            lbincr = solution.lower_bound_incr_add(c)
            if lbincr is None:
                raise ValueError("lbincr cannot be None")
            if isclose(lbincr, 0.0):
                cszero.append(c)
            else:
                cs.append(c)
                p.append((tau[k]**alpha) * ((1.0 / lbincr) ** beta))
        if best is None:
            if len(cszero) > 0:
                best = random.choice(cszero)
            elif len(cs) > 0:
                if sum(p) > 0:
                    best = random.choices(cs, p, k = 1)[0]
                else:
                    best = random.choice(cs)
            else:
                break
        solution.add(best)
    return solution

def ant_system(solutions: list[Solution],
               budget: float,
               tau0: float,
               alpha: float = 1.0,
               beta: float = 3.0,
               rho: float = 0.5,
               seed: int | None = None,
               local_search: LocalSearch[Solution] | None = None,
               ) -> Solution | None:
    start = perf_counter()

    if seed is not None:
        random.seed(seed)

    tau: dict[ComponentId, float] = defaultdict(lambda: tau0)
    nants = len(solutions)

    best, bestobj = None, None
    nc = 0
    while perf_counter() - start < budget:
        nc += 1
        # Build ants
        logging.debug(f"Building ants (iteration {nc})")
        ants = []
        for s in solutions:
            if perf_counter() - start >= budget:
                break
            ant = construct_ant(s.copy(), alpha, beta, tau)

            if ant.is_feasible() and (bestobj is None or ant.objective() < bestobj):
                best = ant.copy()
                bestobj = cast(ObjectiveBound, ant.objective())
                logging.info(f"New best solution found (iteration {nc}): {bestobj}")

            if local_search is not None:
                ant = local_search(ant)
                if ant.is_feasible() and (bestobj is None or cast(ObjectiveBound, ant.objective()) < bestobj):
                    best = ant.copy()
                    bestobj = cast(ObjectiveBound, ant.objective())
                    logging.info(f"New best solution found (iteration {nc}): {bestobj}")
            ants.append(ant)

        if len(ants) > 0:
            mi = math.inf
            ma = -math.inf
            su = 0.0
            c = 0
            for ant in ants:
                if ant.is_feasible():
                    obj = cast(ObjectiveBound, ant.objective())
                    mi = min(mi, obj)
                    ma = max(ma, obj)
                    su += obj
                    c += 1
            avg = su / c if c > 0 else None
            logging.debug(f"Minimum (iteration {nc}): {mi}")
            logging.debug(f"Average (iteration {nc}): {avg}")
            logging.debug(f"Maximum (iteration {nc}): {ma}")

        # Update pheromones
        logging.debug(f"Updating pheromones (iteraton {nc})")
        tau0 = (1.0 - rho) * tau0
        cs = set()
        for k in tau:
            tau[k] = (1.0 - rho) * tau[k]
        for ant in ants:
            if ant.is_feasible():
                obj = cast(ObjectiveBound, ant.objective())
                for comp in ant.components():
                    cid = comp.cid
                    if cid not in tau:
                        tau[cid] = tau0 + 1 / obj
                    else:
                        tau[cid] += 1 / obj
                    cs.add(cid)
            else:
                logging.warn("Ant is not feasible")
        logging.debug(f"Unique components (iteration {nc}): {len(cs)}")
        logging.debug(f"Best solution so far (iteration {nc}): {bestobj}")
    logging.info(f"Number of iterations: {nc}")
    return best
