import random, operator, time, math
import bisect

import logging
logger = logging.getLogger("solvers")

from time import perf_counter
from typing import cast
from operator import itemgetter

from utils.types import Solution, Value, Move, Increment
from roas.greedy_construction import greedy_construction

BSList = list[tuple[Value, Solution]]


import logging
logger = logging.getLogger("solvers")


class KMin:
    """
    Class to keep a set of the k min objects according to a key
    function
    """
    def __init__(self, k, key):
        self.k = k
        self.key = key
        self.keys = []
        self.values = []

    def insert(self, value):
        if len(self.values) == self.k:
            key = self.key(value)
            if key < self.keys[-1]:
                i = bisect.bisect_right(self.keys, key)
                self.keys.insert(i, key)
                self.keys.pop()
                self.values.insert(i, value)
                self.values.pop()
        else:
            self.keys.append(self.key(value))
            self.values.append(value)

    def __str__(self):
        return f"KMin({self.k}, {self.key}, {self.keys}, {self.values})"

    def __iter__(self):
        return self.values.__iter__()

    def __len__(self):
        return self.values.__len__()

def candidates_lookahead(prev: BSList, bw: int) -> KMin:
    result = KMin(bw, key=itemgetter(0))
    for lb, s in prev:
        move_iter = iter(s.construction_neighbourhood().moves(s))
        for c in move_iter:
            result.insert((lb+cast(Value, c.lower_bound_increment(s)), s, c))
    return result

def candidates_full_rollout(prev: BSList, bw: int, bf: int) -> KMin:
    result = KMin(bw, key=itemgetter(0))
    for lb, s in prev:
        move_iter = iter(s.construction_neighbourhood().moves(s))
        for _ in range(bf): 
            c = next(move_iter, None)
            if c is None:
                break
            ns = s.copy()
            c.apply(ns)
            ns = greedy_construction(ns)
            result.insert(( cast(Value, ns.objective_value()), s, c)) # TODO: make more efficient
    return result

def evolve(candidates: KMin) -> BSList:
    result: BSList = []
    for lb, s, c in candidates:
        ns = s.copy()
        c.apply(ns)
        result.append((lb, ns))
    return result

def beam_search(solution: Solution, bw: int = 10, FullRollout: bool = True, bf: int = 5) -> Solution:
    if FullRollout:
        candidates = candidates_full_rollout
    else:
        candidates = candidates_lookahead
    best = solution
    bestobj = best.objective_value()
    v = [(cast(Value, solution.lower_bound()), solution)]
    i = 0
    while True:
        i += 1
        logger.debug(f"Iteration {i}")        
        c = candidates(v, bw, bf)
        if len(c) == 0:
            break
        v = evolve(c)
        for _, s in v:
            if s.is_feasible():
                obj = cast(Value, s.objective_value())
                if bestobj is None or obj < bestobj:
                    best, bestobj = s, obj
                    logger.debug(f"New best solution found: {bestobj}")
    return best
