
import random, operator, math

from time import perf_counter
from typing import cast
from utils.types import Solution, Value, Move, Increment, Neighbourhood
from utils.utils import isclose

import logging
logger = logging.getLogger("solvers")

def rls(solution: Solution, budget: float, prob: float=0.1) -> Solution:
    "Random local search"
    local_nbhood = solution.local_neighbourhood()
    best = solution.copy()
    bestobj = best.objective_value()
    start = perf_counter()
    while perf_counter() - start < budget:
        if random.random() < prob:
            # Randomly select a move from the neighbourhood
            move = local_nbhood.random_move(solution)
            incr = move.objective_value_increment(solution)
            move.apply(solution)
            if solution.value < bestobj:
                best = solution.copy()
                bestobj = solution.value
            break
        else:
            # Select the best move from the neighbourhood
            move_iter = iter(local_nbhood.moves(solution))
            best_move, best_incr = None, math.inf
            for move in move_iter:
                incr = move.objective_value_increment(solution)
                if incr < best_incr:
                    best_move, best_incr = move, incr
            if best_move is not None:
                best_move.apply(solution)
                logger.debug(f"new sol found: {best_incr} {solution.value}")
                if solution.value < bestobj:
                    best = solution.copy()
                    bestobj = solution.value

    return best



def ls_with_side_walks(solution: Solution, budget: float) -> Solution:
    "best improvement search with side moves"
    local_nbhood = solution.local_neighbourhood()
    start = perf_counter()
    while perf_counter() - start < budget:
        move_iter = iter(local_nbhood.random_moves_without_replacement(solution))
        best_move, best_incr = None, math.inf
        for move in move_iter:
            delta = move.objective_value_increment(solution)
            if delta < best_incr:
                best_move, best_incr = move, delta
        if best_move is not None:
            if best_incr < 0 or isclose(delta, 0.0):
                logger.debug(f"new sol found: {delta} {solution.value}")
                move.apply(solution)
            else:
                break
        if perf_counter() - start >= budget:
            return solution

    return solution
