import random, operator, math

from collections.abc import Callable, Iterable
from typing import cast
from utils.types import Solution, Value, Move, Increment
from utils.utils import isclose

from time import perf_counter


import logging
logger = logging.getLogger("solvers")

class LinearDecay:
    def __init__(self, init_temp: float) -> None:
        self.init_temp = init_temp

    def __call__(self, t: float) -> float:
        return t * self.init_temp

class ExponentialAcceptance:
    def __init__(self) -> None:
        ...

    def __call__(self, incr: float, t: float) -> float:
        if incr <= 0:
            return 1.0
        else:
            return math.exp(-incr / t)

def sa(solution: Solution,
       budget: float,
       init_temp: float,
       seed: int | None = None,
       temperature: Callable[[float], float] | None = None,
       acceptance: Callable[[float, float], float] | None = None,
       ) -> Solution:
    if seed is not None:
        random.seed(seed)
        
    if temperature is None:
        temperature = LinearDecay(init_temp)

    if acceptance is None:
        acceptance = ExponentialAcceptance()
       
    start = perf_counter()
    best = solution.copy()
    bestobj = best.objective_value()
    local_nbhood = solution.local_neighbourhood()

    while perf_counter() - start < budget:
        move_iter = iter(local_nbhood.random_moves_without_replacement(solution))
        for move in move_iter:
            t = temperature(1 - (perf_counter()-start) / budget)
            if t <= 0:
                break
            incr = move.objective_value_increment(solution)
            
            if acceptance(incr, t) >= random.random():
                move.apply(solution)                
                #logging.debug(f"new sol found: {incr} {solution.value}")
                if bestobj is None or solution.value < bestobj:
                    best = solution.copy()
                    bestobj = solution.value
                    logger.debug(f"New best solution: {bestobj}")
                break
    return best
