import random, operator, math
import time

import logging
logger = logging.getLogger("solvers")

from utils.types import Solution, Value, Neighbourhood, Move, Increment


def is_tabu(move: Move, tabu_map: dict[Move, int], current_time: int) -> bool:
    """
    Check if a move is in the tabu list.
    """
    if move in tabu_map:
        return current_time < tabu_map[move]
    return False

def aspiration(incr: Increment, value: Value, best_obj: Value) -> bool:
    """
    Check if a move is an aspiration move.
    """
    if incr+value < best_obj:
        # If the move is better than the best solution found so far, it is an aspiration move
        return True
    return False

def tabu_tenure(interval: tuple[int,int] = (5,10)) -> int:
    """
    Calculate the tabu tenure for a move.
    """
    return random.randint(interval[0],interval[1]+1)


def tabu_search(solution: Solution, tabu_interval: tuple[int, int] = (5, 10), time_budget: int = math.inf, max_no_improv: int = math.inf) -> Solution:
    start = time.perf_counter()
    local_nbhood = solution.local_neighbourhood()

    tabu_map: dict[Move, int] = {}

    best = solution.copy()
    best_obj = best.objective_value()
    idle_iterations = 0
    iterations = 0
    
    while time.perf_counter() - start < time_budget and idle_iterations < max_no_improv:
        iterations += 1
        move_iterator = iter(local_nbhood.moves(solution))
        best_move, best_incr = None, math.inf
        for move in move_iterator:
            incr = move.objective_value_increment(solution)
            # Check if the move is in the tabu list
            if is_tabu(move, tabu_map, iterations) and not aspiration(incr, solution.value, best_obj):
                continue
            if incr < best_incr:
                best_move, best_incr = move, incr        
        if best_move is not None:
            best_move.apply(solution)
            logger.debug(f"new sol found in ts: {best_incr} {solution.objective_value()}")
            tabu_map[best_move] = tabu_tenure(tabu_interval) + iterations
            tabu_map[best_move.invert_move()] = tabu_tenure(tabu_interval) + iterations
            if solution.value < best_obj:
                best = solution.copy()
                best_obj = solution.value
                idle_iterations = 0
            else:
                idle_iterations += 1
        else:
            # No move found, the tabu list is too strong
            # jump in iterations 
            iterations += 1 # FIXME find more elegant way to do this
    
    return best
